@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2011-2015  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2011-2019  Ruby-GNOME Project Team
  *  Copyright (C) 2002,2003  Masahiro Sakai
  *
  *  This library is free software; you can redistribute it and/or
@@ -220,7 +220,7 @@ rbg_cstr2rval_with_free(gchar *str)
     return rbg_cstr2rval_free(str);
 }
 
-static rb_encoding *filename_encoding_if_not_utf8;
+rb_encoding *rbg_filename_encoding;
 
 static VALUE
 rbg_filename_to_ruby_body(VALUE filename)
@@ -234,9 +234,9 @@ rbg_filename_to_ruby_body(VALUE filename)
 
     /* if needed, change encoding of Ruby String to filename encoding, so that
        upcoming File operations will work properly */
-    return filename_encoding_if_not_utf8 != NULL ?
-        rb_str_export_to_enc(rb_filename, filename_encoding_if_not_utf8) :
-        rb_filename;
+    return rbg_filename_encoding == rb_utf8_encoding() ?
+        rb_filename :
+        rb_str_export_to_enc(rb_filename, rbg_filename_encoding);
 }
 
 static VALUE
@@ -257,7 +257,7 @@ rbg_filename_to_ruby(const gchar *filename)
     if (filename == NULL)
         return Qnil;
 
-    if (filename_encoding_if_not_utf8 == NULL)
+    if (rbg_filename_encoding == rb_utf8_encoding())
         return CSTR2RVAL(filename);
 
     filename_utf8 = g_filename_to_utf8(filename, -1, NULL, &written, &error);
@@ -278,15 +278,15 @@ rbg_filename_to_ruby_free(gchar *filename)
         return Qnil;
 
     /* convert filename to UTF-8 if needed */
-    if (filename_encoding_if_not_utf8 != NULL) {
+    if (rbg_filename_encoding == rb_utf8_encoding()) {
+        filename_utf8 = filename;
+    } else {
         GError *error = NULL;
 
         filename_utf8 = g_filename_to_utf8(filename, -1, NULL, &written, &error);
         g_free(filename);
         if (error != NULL)
             RAISE_GERROR(error);
-    } else {
-        filename_utf8 = filename;
     }
 
     return rb_ensure(rbg_filename_to_ruby_body, (VALUE)filename_utf8,
@@ -306,10 +306,14 @@ rbg_filename_from_ruby(VALUE filename)
         filename = rb_str_export_to_enc(filename, rb_utf8_encoding());
 
     /* convert it to filename encoding if needed */
-    if (filename_encoding_if_not_utf8 == NULL)
-        return g_strdup(RSTRING_PTR(filename));
+    if (rbg_filename_encoding == rb_utf8_encoding())
+        return g_strndup(RSTRING_PTR(filename), RSTRING_LEN(filename));
 
-    retval = g_filename_from_utf8(RSTRING_PTR(filename), -1, NULL, &written, &error);
+    retval = g_filename_from_utf8(RSTRING_PTR(filename),
+                                  RSTRING_LEN(filename),
+                                  NULL,
+                                  &written,
+                                  &error);
     if (error != NULL)
         RAISE_GERROR(error);
 
@@ -319,7 +323,7 @@ rbg_filename_from_ruby(VALUE filename)
 struct rval2strv_args {
     VALUE ary;
     long n;
-    const gchar **result;
+    gchar **result;
 };
 
 static VALUE
@@ -329,28 +333,28 @@ rbg_rval2strv_body(VALUE value)
     struct rval2strv_args *args = (struct rval2strv_args *)value;
 
     for (i = 0; i < args->n; i++)
-        args->result[i] = RVAL2CSTR(RARRAY_PTR(args->ary)[i]);
+        args->result[i] = (gchar *)RVAL2CSTR(RARRAY_PTR(args->ary)[i]);
     args->result[args->n] = NULL;
 
     return Qnil;
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2strv_rescue(VALUE value)
+rbg_rval2strv_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rval2strv_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
-const gchar **
+gchar **
 rbg_rval2strv(volatile VALUE *value, long *n)
 {
     struct rval2strv_args args;
 
     args.ary = *value = rb_ary_dup(rb_ary_to_ary(*value));
     args.n = RARRAY_LEN(args.ary);
-    args.result = g_new(const gchar *, args.n + 1);
+    args.result = g_new(gchar *, args.n + 1);
 
     rb_rescue(rbg_rval2strv_body, (VALUE)&args,
               rbg_rval2strv_rescue, (VALUE)&args);
@@ -361,7 +365,7 @@ rbg_rval2strv(volatile VALUE *value, long *n)
     return args.result;
 }
 
-const gchar **
+gchar **
 rbg_rval2strv_accept_nil(volatile VALUE *value, long *n)
 {
     if (!NIL_P(*value))
@@ -393,11 +397,11 @@ rbg_rval2strv_dup_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2strv_dup_rescue(VALUE value)
+rbg_rval2strv_dup_rescue(VALUE value, VALUE error)
 {
-    g_free(((struct rval2strv_dup_args *)value)->result);
+    g_strfreev(((struct rval2strv_dup_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 gchar **
@@ -407,7 +411,7 @@ rbg_rval2strv_dup(volatile VALUE *value, long *n)
 
     args.ary = *value = rb_ary_dup(rb_ary_to_ary(*value));
     args.n = RARRAY_LEN(args.ary);
-    args.result = g_new(gchar *, args.n + 1);
+    args.result = g_new0(gchar *, args.n + 1);
 
     rb_rescue(rbg_rval2strv_dup_body, (VALUE)&args,
               rbg_rval2strv_dup_rescue, (VALUE)&args);
@@ -430,6 +434,63 @@ rbg_rval2strv_dup_accept_nil(volatile VALUE *value, long *n)
     return NULL;
 }
 
+struct rval2filenamev_args {
+    VALUE ary;
+    long n;
+    gchar **result;
+};
+
+static VALUE
+rbg_rval2filenamev_body(VALUE value)
+{
+    long i;
+    struct rval2filenamev_args *args = (struct rval2filenamev_args *)value;
+
+    for (i = 0; i < args->n; i++)
+        args->result[i] = RVAL2CSTRFILENAME(RARRAY_PTR(args->ary)[i]);
+    args->result[args->n] = NULL;
+
+    return Qnil;
+}
+
+static G_GNUC_NORETURN VALUE
+rbg_rval2filenamev_rescue(VALUE value, VALUE error)
+{
+    g_strfreev(((struct rval2filenamev_args *)value)->result);
+
+    rb_exc_raise(error);
+}
+
+gchar **
+rbg_rval2filenamev(volatile VALUE *value, long *n)
+{
+    struct rval2filenamev_args args;
+
+    args.ary = *value = rb_ary_dup(rb_ary_to_ary(*value));
+    args.n = RARRAY_LEN(args.ary);
+    args.result = g_new0(gchar *, args.n + 1);
+
+    rb_rescue(rbg_rval2filenamev_body, (VALUE)&args,
+              rbg_rval2filenamev_rescue, (VALUE)&args);
+
+    if (n != NULL)
+        *n = args.n;
+
+    return args.result;
+}
+
+gchar **
+rbg_rval2filenamev_accept_nil(volatile VALUE *value, long *n)
+{
+    if (!NIL_P(*value))
+        return rbg_rval2filenamev(value, n);
+
+    if (n != NULL)
+        *n = 0;
+
+    return NULL;
+}
+
 VALUE
 rbg_strv2rval(const gchar **strings)
 {
@@ -486,11 +547,11 @@ rbg_rval2gbooleans_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2gbooleans_rescue(VALUE value)
+rbg_rval2gbooleans_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2gbooleans_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 gboolean *
@@ -529,11 +590,11 @@ rbg_rval2gints_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2gints_rescue(VALUE value)
+rbg_rval2gints_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2gints_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 gint *
@@ -572,11 +633,11 @@ rbg_rval2gint8s_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2gint8s_rescue(VALUE value)
+rbg_rval2gint8s_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2gint8s_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 gint8 *
@@ -615,11 +676,11 @@ rbg_rval2guint8s_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2guint8s_rescue(VALUE value)
+rbg_rval2guint8s_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2guint8s_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 guint8 *
@@ -658,11 +719,11 @@ rbg_rval2guint16s_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2guint16s_rescue(VALUE value)
+rbg_rval2guint16s_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2guint16s_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 guint16 *
@@ -701,11 +762,11 @@ rbg_rval2guint32s_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2guint32s_rescue(VALUE value)
+rbg_rval2guint32s_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2guint32s_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 guint32 *
@@ -744,11 +805,11 @@ rbg_rval2gdoubles_body(VALUE value)
 }
 
 static G_GNUC_NORETURN VALUE
-rbg_rval2gdoubles_rescue(VALUE value)
+rbg_rval2gdoubles_rescue(VALUE value, VALUE error)
 {
     g_free(((struct rbg_rval2gdoubles_args *)value)->result);
 
-    rb_exc_raise(rb_errinfo());
+    rb_exc_raise(error);
 }
 
 gdouble *
@@ -854,47 +915,79 @@ rbg_check_hash_type (VALUE object)
 void
 rbg_scan_options (VALUE options, ...)
 {
-    VALUE original_options = options;
-    VALUE available_keys;
     const char *key;
-    VALUE *value;
+    gsize n_keys = 0;
+    gsize n_found_keys = 0;
     va_list args;
+    va_list args_copy;
 
-    if (NIL_P(options)) {
-        options = rb_hash_new();
-    } else {
+    if (!NIL_P(options)) {
+        VALUE original_options = options;
         options = rbg_check_hash_type(options);
-        if (options == original_options) {
-            options = rb_funcall(options, rb_intern("dup"), 0);
-        } else if (NIL_P(options)) {
+        if (NIL_P(options)) {
             rb_raise(rb_eArgError,
                      "options must be Hash or nil: %+" PRIsVALUE,
                      original_options);
         }
     }
 
-    available_keys = rb_ary_new();
     va_start(args, options);
+    va_copy(args_copy, args);
     key = va_arg(args, const char *);
+    n_keys = 0;
+    n_found_keys = 0;
     while (key) {
-        VALUE rb_key;
-        value = va_arg(args, VALUE *);
+        VALUE *value;
 
-        rb_key = ID2SYM(rb_intern(key));
-        rb_ary_push(available_keys, rb_key);
-        *value = rb_funcall(options, rb_intern("delete"), 1, rb_key);
+        value = va_arg(args, VALUE *);
+        if (NIL_P(options)) {
+            *value = Qnil;
+        } else {
+            VALUE rb_key;
+            rb_key = ID2SYM(rb_intern(key));
+            if (RTEST(rb_funcall(options, rb_intern("key?"), 1, rb_key))) {
+                n_found_keys++;
+            }
+            *value = rb_hash_aref(options, rb_key);
+        }
+        n_keys++;
 
         key = va_arg(args, const char *);
     }
     va_end(args);
 
-    if (RVAL2CBOOL(rb_funcall(options, rb_intern("empty?"), 0)))
+    if (NIL_P(options)) {
         return;
+    }
 
-    rb_raise(rb_eArgError,
-             "unexpected key(s) exist: %s: available keys: %s",
-             rbg_inspect(rb_funcall(options, rb_intern("keys"), 0)),
-             rbg_inspect(available_keys));
+    if (n_found_keys == RHASH_SIZE(options)) {
+        return;
+    }
+
+    {
+        VALUE rb_available_keys;
+
+        rb_available_keys = rb_ary_new();
+        key = va_arg(args_copy, const char *);
+        while (key) {
+            VALUE rb_key;
+
+            va_arg(args_copy, VALUE *);
+            rb_key = ID2SYM(rb_intern(key));
+            rb_ary_push(rb_available_keys, rb_key);
+            key = va_arg(args_copy, const char *);
+        }
+        va_end(args_copy);
+
+        rb_raise(rb_eArgError,
+                 "unexpected key(s) exist: %+" PRIsVALUE
+                 ": available keys: %+" PRIsVALUE,
+                 rb_funcall(rb_funcall(options, rb_intern("keys"), 0),
+                            rb_intern("-"),
+                            1,
+                            rb_available_keys),
+                 rb_available_keys);
+    }
 }
 
 #if 0
@@ -979,6 +1072,10 @@ extern void Init_glib2(void);
 void
 Init_glib2(void)
 {
+#ifdef HAVE_RB_EXT_RACTOR_SAFE
+    rb_ext_ractor_safe(true);
+#endif
+
     const gchar **filename_charsets;
 
     id_inspect = rb_intern("inspect");
@@ -1023,9 +1120,9 @@ Init_glib2(void)
     rb_define_const(RG_TARGET_NAMESPACE, "MAXSHORT", INT2FIX(G_MAXSHORT));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXUSHORT", UINT2NUM(G_MAXUSHORT));
 
-    rb_define_const(RG_TARGET_NAMESPACE, "MINLONG", INT2FIX(G_MINLONG));
-    rb_define_const(RG_TARGET_NAMESPACE, "MAXLONG", INT2NUM(G_MAXLONG));
-    rb_define_const(RG_TARGET_NAMESPACE, "MAXULONG", UINT2NUM(G_MAXULONG));
+    rb_define_const(RG_TARGET_NAMESPACE, "MINLONG", LL2NUM(G_MINLONG));
+    rb_define_const(RG_TARGET_NAMESPACE, "MAXLONG", LL2NUM(G_MAXLONG));
+    rb_define_const(RG_TARGET_NAMESPACE, "MAXULONG", ULL2NUM(G_MAXULONG));
 
     rb_define_const(RG_TARGET_NAMESPACE, "MININT8", INT2FIX(G_MININT8));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXINT8", INT2FIX(G_MAXINT8));
@@ -1039,16 +1136,16 @@ Init_glib2(void)
     rb_define_const(RG_TARGET_NAMESPACE, "MAXINT32", INT2NUM(G_MAXINT32));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXUINT32", UINT2NUM(G_MAXUINT32));
 
-    rb_define_const(RG_TARGET_NAMESPACE, "MININT64", INT2FIX(G_MININT64));
+    rb_define_const(RG_TARGET_NAMESPACE, "MININT64", LL2NUM(G_MININT64));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXINT64", LL2NUM(G_MAXINT64));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXUINT64", ULL2NUM(G_MAXUINT64));
 
-    rb_define_const(RG_TARGET_NAMESPACE, "MAXSIZE", UINT2NUM(G_MAXSIZE));
+    rb_define_const(RG_TARGET_NAMESPACE, "MAXSIZE", ULL2NUM(G_MAXSIZE));
 
-    rb_define_const(RG_TARGET_NAMESPACE, "MINFLOAT", INT2FIX(G_MINFLOAT));
+    rb_define_const(RG_TARGET_NAMESPACE, "MINFLOAT", DBL2NUM(G_MINFLOAT));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXFLOAT", DBL2NUM(G_MAXFLOAT));
 
-    rb_define_const(RG_TARGET_NAMESPACE, "MINDOUBLE", INT2FIX(G_MINDOUBLE));
+    rb_define_const(RG_TARGET_NAMESPACE, "MINDOUBLE", DBL2NUM(G_MINDOUBLE));
     rb_define_const(RG_TARGET_NAMESPACE, "MAXDOUBLE", DBL2NUM(G_MAXDOUBLE));
 
     /* Standard Macros */
@@ -1065,12 +1162,13 @@ Init_glib2(void)
         || filename_charsets[0] == NULL
         || !strcmp(filename_charsets[0], "UTF-8")
         || rb_enc_find(filename_charsets[0]) == rb_enc_find("ASCII-8BIT")) {
-        /* set to NULL, mean do not perform transcoding, either filename
-           encoding is unknown, UTF-8, or unsupported */
-        filename_encoding_if_not_utf8 = NULL;
+        rbg_filename_encoding = rb_utf8_encoding();
     } else {
-        filename_encoding_if_not_utf8 = rb_enc_find(filename_charsets[0]);
+        rbg_filename_encoding = rb_enc_find(filename_charsets[0]);
     }
+    rb_define_const(RG_TARGET_NAMESPACE,
+                    "FILENAME_ENCODING",
+                    rb_enc_from_encoding(rbg_filename_encoding));
 
 /* Don't implement them.
 #define     G_DIR_SEPARATOR_S
@@ -1114,6 +1212,9 @@ union       GDoubleIEEE754;
     rb_define_const(RG_TARGET_NAMESPACE, "PRIORITY_LOW", INT2FIX(G_PRIORITY_LOW));
 
 /*    Init_mem(); */
+
+    Init_glib_gc();
+
     Init_gutil();
     Init_gutil_callback();
 
@@ -1130,6 +1231,7 @@ union       GDoubleIEEE754;
     Init_gobject();
 
     /* Require GBoxed/GObject */
+    Init_glib_error_conversions();
     Init_glib_utils();
     Init_glib_spawn();
     Init_glib_spawnerror();
@@ -1155,4 +1257,7 @@ union       GDoubleIEEE754;
     Init_glib_variant();
     Init_glib_regex();
     Init_glib_matchinfo();
+    Init_glib_date_time();
+    Init_glib_time_zone();
+    Init_glib_bytes();
 }
